package executor;

import algorithms.Classifier;
import algorithms.MotionLoaderBoneRotations;
import algorithms.MotionLoaderJointCoordinates;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import objects.PoseJointCoordinates;
import objects.MotionJointCoordinates;
import objects.Feature;
import objects.Motion;
import methods.Method;
import methods.MethodAhmed;
import methods.MethodAli;
import methods.MethodAndersson;
import methods.MethodBall;
import methods.MethodDikovski;
import methods.MethodGavrilova;
import methods.MethodJiang;
import methods.MethodKrzeszowski;
import methods.MethodKumar;
import methods.MethodKwolek;
import methods.MethodPreis;
import methods.MethodSedmidubsky;
import methods.MethodSinha;
import methods.Method_MMC;
import methods.Method_PCALDA;
import methods.Method_Random;
import methods.Method_Raw;
import objects.ClassifiableList;
import objects.Parameter;
import objects.Template;

public class Executor {

    /*
     Select distance threshold
     Each value refers to a particular sub-database
     */
    private static final double distanceThreshold = 302.0; // 56.3(2,35), 59.4(4,67), 63.3(8,130), 73.7(16,302), 173.3(32,2047), 302.0(54,3843), 495.3(64,5923)

    /*
     Select actions
     */
    private static final boolean extractDatabase = false;
    private static final boolean learnClassifiers = false;
    private static final boolean performClassifications = false;
    private static final boolean evaluateMethods = true;

    /*
     Select evaluation set-ups at evaluateMethods()
     */
    private static final boolean evaluateClassification = true;
    private static final boolean evaluateClassSeparability = false;

    /*
     Motion loaders
     */
    private static final MotionLoaderBoneRotations motionLoaderBoneRotations = new MotionLoaderBoneRotations();
    private static final MotionLoaderJointCoordinates motionLoaderJointCoordinates = new MotionLoaderJointCoordinates();

    /*
     All implemented MoCap-based gait recognition methods
     */
    private static final Method[] methods = new Method[]{
        new MethodAhmed(motionLoaderJointCoordinates),//0
        new MethodAli(motionLoaderJointCoordinates),//1
        new MethodAndersson(motionLoaderJointCoordinates),//2
        new MethodBall(motionLoaderJointCoordinates),//3
        new MethodDikovski(motionLoaderJointCoordinates),//4
        new MethodGavrilova(motionLoaderJointCoordinates),//5
        new MethodJiang(motionLoaderJointCoordinates),//6
        new MethodKrzeszowski(motionLoaderJointCoordinates),//7
        new MethodKumar(motionLoaderJointCoordinates),//8
        new MethodKwolek(motionLoaderJointCoordinates),//9
        new MethodPreis(motionLoaderJointCoordinates),//10
        new MethodSedmidubsky(motionLoaderJointCoordinates),//11
        new MethodSinha(motionLoaderJointCoordinates),//12
        new Method_MMC(motionLoaderBoneRotations),//13
        new Method_MMC(motionLoaderJointCoordinates),//14
        new Method_PCALDA(motionLoaderBoneRotations),//15
        new Method_PCALDA(motionLoaderJointCoordinates),//16
        new Method_Random(motionLoaderBoneRotations),//17
        new Method_Random(motionLoaderJointCoordinates),//18
        new Method_Raw(motionLoaderBoneRotations),//19
        new Method_Raw(motionLoaderJointCoordinates),//20
    };

    /*
     Main method
     Executes the actions selected above
     */
    public static void main(String[] args) throws IOException, ClassNotFoundException {
        if (extractDatabase) {
            extractDatabase();
        }
        if (learnClassifiers) {
            learnClassifiers();
        }
        if (performClassifications) {
            performClassifications();
        }
        if (evaluateMethods) {
            evaluateMethods(args);
        }
    }

    /*
     Normalize the whole original CMU database in amcOriginal/ directory with respect to person’s position and walk direction and saves it to amc/ directory
     Select gait cycles as sub-motions with similarity under selected distanceThreshold to prototypicalFileAMC and save them to separate AMC files {subject}_{id}[{from}-{to}].amc in amc{distanceThreshold}/ directory
     Only subjects of at least 10 gait cycles are kept.
     */
    public static void extractDatabase() throws IOException {
        new File("amc/").mkdirs();
        for (File fileAMC : new File("amcOriginal/").listFiles()) {
            BufferedReader brAMC = new BufferedReader(new FileReader(fileAMC));
            BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc/" + fileAMC.getName()));
            String lineAMC = brAMC.readLine();
            while (lineAMC != null) {
                if (lineAMC.split(" ")[0].equals("root")) {
                    wrAMC.write("root 0 0 0 0 0 0\r\n");
                } else {
                    wrAMC.write(lineAMC + "\r\n");
                }
                lineAMC = brAMC.readLine();
            }
            wrAMC.flush();
            wrAMC.close();
        }
        MotionJointCoordinates motionProto = new MotionLoaderJointCoordinates().loadMotion(Constants.prototypicalFileAMC);
        Feature featureProto = motionProto.extractFeetDistanceSideSensitiveFeature();
        int lengthProto = featureProto.getLength();
        double sigma = 2;
        int minLength = (int) Math.round(lengthProto / sigma);
        int maxLength = (int) Math.round(lengthProto * sigma);
        List<Motion> gaitCycles = new ArrayList();
        for (File fileAMC : new File("amc/").listFiles()) {
            System.out.print("Processing " + fileAMC.getName() + "\r\n");
            MotionJointCoordinates motion = new MotionLoaderJointCoordinates().loadMotion(fileAMC);
            Feature feature = motion.extractFeetDistanceSideSensitiveFeature();
            int length = feature.getLength();
            boolean selected = true;
            while (selected) {
                selected = false;
                double minDistance = Double.MAX_VALUE;
                int minFrom = 0;
                int minTo = 0;
                for (int from = 0; from < length; from++) {
                    for (int to = from + minLength; to < Math.min(from + maxLength, length); to++) {
                        Feature subFeature = feature.getSubFeature(from, to);
                        subFeature.adjust(lengthProto);
                        double[] subFeatureParameterValues = subFeature.getParameterValues();
                        double[] featureProtoParameterValues = featureProto.getParameterValues();
                        double distance = 0;
                        for (int l = 0; l < lengthProto; l++) {
                            distance += Math.abs(subFeatureParameterValues[l] - featureProtoParameterValues[l]);
                        }
                        if (minDistance > distance) {
                            minDistance = distance;
                            minFrom = from;
                            minTo = to;
                        }
                    }
                }
                if (minDistance < distanceThreshold) {
                    List<PoseJointCoordinates> subPoses = new ArrayList();
                    for (int l = minFrom; l < minTo && l < motion.getLength(); l++) {
                        subPoses.add((PoseJointCoordinates) motion.getPose(l));
                    }
                    gaitCycles.add(new MotionJointCoordinates(motion.getSubject() + "_" + motion.getSequence() + "_" + (minFrom + 1) + "-" + (minTo + 1), subPoses));
                    selected = true;
                    for (int l = minFrom; l < minTo; l++) {
                        feature.setParameter(l, new Parameter("Erased", new Random().nextDouble() * Double.MAX_VALUE));
                    }
                }
            }
        }
        List<List<Motion>> gaitCyclesByClass = new ClassifiableList(gaitCycles).splitBySubject();
        new File("amc" + distanceThreshold + "/").mkdirs();
        for (List<Motion> gaitCyclesOfClass : gaitCyclesByClass) {
            if (gaitCyclesOfClass.size() >= 10) {
                for (Motion gaitCycle : gaitCyclesOfClass) {
                    String identifier = gaitCycle.getIdentifier();
                    BufferedReader brAMC = new BufferedReader(new FileReader("amc/" + gaitCycle.getSubject() + "_" + gaitCycle.getSequence() + ".amc"));
                    BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc" + distanceThreshold + "/" + identifier + ".amc"));
                    String subSequence = identifier.split("_")[2];
                    String from = subSequence.split("-")[0];
                    String to = subSequence.split("-")[1];
                    String lineAMC = brAMC.readLine();
                    while (lineAMC != null && !lineAMC.equals("1")) {
                        wrAMC.write(lineAMC + "\r\n");
                        lineAMC = brAMC.readLine();
                    }
                    while (lineAMC != null && !lineAMC.equals(String.valueOf(from))) {
                        lineAMC = brAMC.readLine();
                    }
                    int f = 0;
                    while (lineAMC != null && !lineAMC.equals(to)) {
                        if (lineAMC.split(" ").length == 1) {
                            wrAMC.write(++f + "\r\n");
                        } else {
                            wrAMC.write(lineAMC + "\r\n");
                        }
                        lineAMC = brAMC.readLine();
                    }
                    wrAMC.flush();
                    wrAMC.close();
                    brAMC.close();
                }
            }
        }
    }

    /*
     Extracts classifiers for all methods learned on the sub-database determined by distanceThreshold
     Classifiers are saved as separate files {methodName}-{distanceThreshold}.classifier in classifiers/ directory
     */
    private static void learnClassifiers() throws IOException {
        new File("classifiers/").mkdirs();
        for (Method method : methods) {
            method.learnClassifier(method.loadMotions(new File("amc" + distanceThreshold + "/")));
            method.saveClassifier(new File("classifiers/" + method.getName() + "-" + distanceThreshold + ".classifier"));
        }
    }

    /*
     Performs classifications of customProbeFileAMC over customGalleryDirectory for all methods
     Instead of learning classifiers methods load their extracted classifiers {distanceThreshold}-{methodName}.classifier in classifiers/ directory that are learned on the sub-database determined by distanceThreshold
     */
    private static void performClassifications() throws IOException, ClassNotFoundException {
        ObjectInputStream oos = new ObjectInputStream(new FileInputStream(Constants.customClassifier));
        Classifier classifier = (Classifier) oos.readObject();
        oos.close();
        for (Method method : methods) {
            classifier.importTemplatesGallery(method.extractTemplates(method.loadMotions(Constants.customGalleryDirectory)));
            Template templateQuery = method.extractTemplate(method.loadMotion(Constants.customProbeFileAMC));
            System.out.print(method.getName() + ": query subject classified as " + classifier.classify(templateQuery) + "\r\n");
        }
    }

    /*
     Performs evaluations for all methods
     */
    private static void evaluateMethods(String[] args) throws IOException {
        long beginningTotal = System.currentTimeMillis();
//        for (Method method : methods) {//
        Method method = methods[Integer.parseInt(args[0])];//Integer.parseInt(args[0])
        System.out.print(method.getName() + "," + distanceThreshold + "\r\n");
        List<Motion> motions = method.loadMotions(new File("amc" + distanceThreshold + "/"));//
//        List<Motion> motions = method.loadMotions1(new File("single1.csv"));//
        int numberOfClasses = new ClassifiableList(motions).splitBySubject().size();
        if (evaluateClassification) {
            int nClasses; // number of learning and evaluation classes
            int nFoldsLearning; // number of learning folds
            int attempts; // number of evaluation attempts
            long beginning; // time of beginning

            //Svenda
//            method.evaluateClassification(motions, numberOfClasses, 10, 1, 30);
            //ACM-TOMM
            System.out.print("CCR,EER,AUC,MAP,DCT,TD\r\n");
            method.evaluateClassification(motions, numberOfClasses, 10, 1, 30);

//            // default
//            nClasses = numberOfClasses;
//            nFoldsLearning = 3;
//            attempts = 1;
//            beginning = System.currentTimeMillis();
//            method.evaluateClassification(motions, nClasses, nFoldsLearning, attempts);
//            System.out.print("timeHomogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
            // S+SSPR Figure 3
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClasses = i;
//                    nFoldsLearning = 3;
//                    attempts = 1;
//                    System.out.print("(" + nClasses + "," + nClasses +")\r\n");
//                    beginning = System.currentTimeMillis();
//                    method.evaluateClassification(motions, nClasses, nFoldsLearning, attempts);
//                    System.out.print("timeHomogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
        }
        if (evaluateClassSeparability) {
            int nClassesLearning; // number of learning classes
            int nClassesEvaluation; // number of evaluation classes
            int attempts; // number of evaluation attempts
            long beginning; // time of beginning

//             default
//                nClassesLearning = 2; // number of learning classes
//                nClassesEvaluation = numberOfClasses - 2;
//                attempts = 3;
//                beginning = System.currentTimeMillis();
//                method.evaluateClassSeparability(motions, nClassesLearning, nClassesEvaluation, attempts);
//                System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
            // S+SSPR Figure 3
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = i;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                    method.evaluateClassSeparability(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
            // S+SSPR Figure 4a
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = numberOfClasses / 2;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                    method.evaluateClassSeparability(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
            // S+SSPR Figure 4b
//                for (int i = 2; i <= numberOfClasses - 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = numberOfClasses - i;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation + ")\r\n");
//                    method.evaluateClassSeparability(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
            // PRL/IJCB
            System.out.print(",DBI,DI,SC,FDR,DCT,TD\r\n");
            for (int i = Integer.parseInt(args[1]); i <= numberOfClasses - 2; i++) {
                nClassesLearning = i;
                nClassesEvaluation = numberOfClasses - i;
                attempts = 1;
                beginning = System.currentTimeMillis();
                System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation + "),");
                method.evaluateClassSeparability(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
            }

//            // keys
//            System.out.print("DBI,DI,SC,FDR,CCR,EER,AUC,MAP,DCT,TD\r\n");
//            method.evaluateClassification(motions, 5, 2, 1);
        }
//        }//
        System.out.print("timeTotal=" + printTime(System.currentTimeMillis() - beginningTotal));
    }

//    public static List<Motion> loadMotionsJointCoordinates(File directoryAMC) throws IOException {
//        List<Motion> motionsMJC = new ArrayList();
//        for (File fileAMC : directoryAMC.listFiles()) {
//            motionsMJC.add(convertASFAMCtoMJC(Constants.prototypicalFileASF, fileAMC));
//        }
//        return motionsMJC;
//    }
//    public static List<Motion> loadMotionsBoneRotations(File directoryAMC) throws IOException {
//        List<Motion> motionsMBR = new ArrayList();
//        for (File fileAMC : directoryAMC.listFiles()) {
//            motionsMBR.add(convertAMCtoMBR(fileAMC));
//        }
//        return motionsMBR;
//    }
//    public static Motion loadMotionJointCoordinates(File fileAMC) throws IOException {
//        return convertASFAMCtoMJC(Constants.prototypicalFileASF, fileAMC);
//    }
//    public static Motion loadMotionBoneRotations(File fileAMC) throws IOException {
//        return convertAMCtoMBR(fileAMC);
//    }
    private static String printTime(double time) {
        String unit = "ms";
        if (time >= 1000) {
            time /= 1000;
            unit = "s";
            if (time >= 60) {
                time /= 60;
                unit = "m";
                if (time >= 60) {
                    time /= 60;
                    unit = "h";
                    if (time >= 24) {
                        time /= 24;
                        unit = "d";
                    }
                }
            }
        }
        return String.valueOf((double) Math.round(time * 10) / 10) + unit;
    }
}
